@@ -9,20 +9,6 @@
 
 #include "xklavier_private.h"
 
-#include <X11/extensions/XKBfile.h>
-#include <X11/extensions/XKM.h>
-
-#define RULES_FILE "xfree86"
-
-#define RULES_PATH ( XKB_BASE "/rules/" RULES_FILE )
-
-#define XML_CFG_PATH ( XKB_BASE "/rules/xfree86.xml" )
-
-#define MULTIPLE_LAYOUTS_CHECK_PATH ( XKB_BASE "/symbols/pc/en_US" )
-
-#define XK_XKB_KEYS
-#include <X11/keysymdef.h>
-
 typedef struct _XklConfigRegistry
 {
   xmlDocPtr doc;
@@ -30,30 +16,36 @@ typedef struct _XklConfigRegistry
 }
 XklConfigRegistry;
 
-XkbRF_VarDefsRec _xklVarDefs;
-
 static XklConfigRegistry theRegistry;
 
 static xmlXPathCompExprPtr modelsXPath;
 static xmlXPathCompExprPtr layoutsXPath;
 static xmlXPathCompExprPtr optionGroupsXPath;
 
-static XkbRF_RulesPtr rules;
-static XkbComponentNamesRec componentNames;
-static char *locale;
+#define _XklConfigRegistryIsInitialized() \
+  ( theRegistry.xpathContext != NULL )
 
-static Bool _XklConfigInitialized(  )
+static xmlChar *_XklNodeGetXmlLangAttr( xmlNodePtr nptr )
 {
-  return theRegistry.xpathContext != NULL;
+  if( nptr->properties != NULL &&
+      !strcmp( "lang", nptr->properties[0].name ) &&
+      nptr->properties[0].ns != NULL &&
+      !strcmp( "xml", nptr->properties[0].ns->prefix ) &&
+      nptr->properties[0].children != NULL )
+    return nptr->properties[0].children->content;
+  else
+    return NULL;
 }
 
 static Bool _XklReadConfigItem( xmlNodePtr iptr, XklConfigItemPtr pci )
 {
-  xmlNodePtr nameElement, descElement = NULL, defDescElement =
-    NULL, nptr, ptr;
-  int maxPriority = -1;
+  xmlNodePtr nameElement, descElement = NULL, ntDescElement =
+    NULL, nptr, ptr, shortDescElement = NULL, ntShortDescElement = NULL;
+  int maxDescPriority = -1;
+  int maxShortDescPriority = -1;
 
   *pci->name = 0;
+  *pci->shortDescription = 0;
   *pci->description = 0;
   if( iptr->type != XML_ELEMENT_NODE )
     return False;
@@ -88,29 +80,37 @@ static Bool _XklReadConfigItem( xmlNodePtr iptr, XklConfigItemPtr pci )
   {
     if( nptr->type != XML_TEXT_NODE )
     {
-      if( nptr->properties != NULL &&
-          !strcmp( "lang", nptr->properties[0].name ) &&
-          nptr->properties[0].ns != NULL &&
-          !strcmp( "xml", nptr->properties[0].ns->prefix ) &&
-          nptr->properties[0].children != NULL )
+      xmlChar *lang = _XklNodeGetXmlLangAttr( nptr );
+
+      if( lang != NULL )
       {
-        xmlChar *lang = nptr->properties[0].children->content;
         int priority = _XklGetLanguagePriority( lang );
-        if( priority > maxPriority )    // higher priority
+        if( !strcmp( nptr->name, "description" ) && ( priority > maxDescPriority ) )    // higher priority
         {
           descElement = nptr;
-          maxPriority = priority;
+          maxDescPriority = priority;
+        } else if( !strcmp( nptr->name, "shortDescription" ) && ( priority > maxShortDescPriority ) )   // higher priority
+        {
+          shortDescElement = nptr;
+          maxShortDescPriority = priority;
         }
       } else
       {
-        defDescElement = nptr;
+        if( !strcmp( nptr->name, "description" ) )
+          ntDescElement = nptr;
+        else if( !strcmp( nptr->name, "shortDescription" ) )
+          ntShortDescElement = nptr;
       }
     }
     nptr = nptr->next;
   }
 
+  // if no language-specific description found - use the ones without lang
   if( descElement == NULL )
-    descElement = defDescElement;
+    descElement = ntDescElement;
+
+  if( shortDescElement == NULL )
+    shortDescElement = ntShortDescElement;
 
   //
   // Actually, here we should have some code to find the correct localized description...
@@ -119,10 +119,24 @@ static Bool _XklReadConfigItem( xmlNodePtr iptr, XklConfigItemPtr pci )
   if( nameElement != NULL && nameElement->children != NULL )
     strncat( pci->name, nameElement->children->content,
              XKL_MAX_CI_NAME_LENGTH - 1 );
+
+  if( shortDescElement != NULL && shortDescElement->children != NULL )
+  {
+    char * lsd = _XklLocaleFromUtf8( shortDescElement->children->content );
+    strncat( pci->shortDescription,
+             lsd,
+             XKL_MAX_CI_SHORT_DESC_LENGTH - 1 );
+    free( lsd );
+  }
+
   if( descElement != NULL && descElement->children != NULL )
+  {
+    char * ld = _XklLocaleFromUtf8( descElement->children->content );
     strncat( pci->description,
-             _XklLocaleFromUtf8( descElement->children->content ),
+             ld,
              XKL_MAX_CI_DESC_LENGTH - 1 );
+    free( ld );
+  }
   return True;
 }
 
@@ -150,7 +164,7 @@ static void _XklConfigEnumSimple( xmlXPathCompExprPtr xpathCompExpr,
 {
   xmlXPathObjectPtr xpathObj;
 
-  if( !_XklConfigInitialized(  ) )
+  if( !_XklConfigRegistryIsInitialized(  ) )
     return;
   xpathObj = xmlXPathCompiledEval( xpathCompExpr, theRegistry.xpathContext );
   if( xpathObj != NULL )
@@ -167,7 +181,7 @@ static void _XklConfigEnumDirect( const char *format,
   char xpathExpr[1024];
   xmlXPathObjectPtr xpathObj;
 
-  if( !_XklConfigInitialized(  ) )
+  if( !_XklConfigRegistryIsInitialized(  ) )
     return;
   snprintf( xpathExpr, sizeof xpathExpr, format, value );
   xpathObj = xmlXPathEval( xpathExpr, theRegistry.xpathContext );
@@ -188,7 +202,7 @@ static Bool _XklConfigFindObject( const char *format,
   Bool rv = False;
   char xpathExpr[1024];
 
-  if( !_XklConfigInitialized(  ) )
+  if( !_XklConfigRegistryIsInitialized(  ) )
     return False;
 
   snprintf( xpathExpr, sizeof xpathExpr, format, arg1, ptr->name );
@@ -283,12 +297,11 @@ void _XklConfigRecSplitByComma( char ***array,
   const char *pc = merged;
   char **ppc, *npc;
   *arraySize = 0;
+  *array = NULL;
 
   if( merged == NULL || merged[0] == '\0' )
-  {
-    *arraySize = 0;
     return;
-  }
+
   // first count the elements 
   while( ( npc = strchr( pc, ',' ) ) != NULL )
   {
@@ -307,99 +320,24 @@ void _XklConfigRecSplitByComma( char ***array,
     while( ( npc = strchr( pc, ',' ) ) != NULL )
     {
       int len = npc - pc;
-      *ppc = ( char * ) strndup( pc, len );
+      //*ppc = ( char * ) strndup( pc, len );
+      *ppc = ( char * ) malloc( len + 1 );
+      if ( *ppc != NULL )
+      {
+        strncpy( *ppc, pc, len );
+        (*ppc)[len] = '\0';
+      }
+
       ppc++;
       pc = npc + 1;
     }
 
-    len = npc - pc;
-    *ppc = ( char * ) strndup( pc, len );
-  }
-}
-
-static Bool _XklConfigPrepareBeforeKbd( const XklConfigRecPtr data )
-{
-  memset( &_xklVarDefs, 0, sizeof( _xklVarDefs ) );
-
-  _xklVarDefs.model = ( char * ) data->model;
-
-  if( data->layouts != NULL )
-    _xklVarDefs.layout = _XklConfigRecMergeLayouts( data );
-
-  if( data->variants != NULL )
-    _xklVarDefs.variant = _XklConfigRecMergeVariants( data );
-
-  if( data->options != NULL )
-    _xklVarDefs.options = _XklConfigRecMergeOptions( data );
-
-  locale = setlocale( LC_ALL, NULL );
-  if( locale != NULL )
-    locale = strdup( locale );
-
-  rules = XkbRF_Load( RULES_PATH, locale, True, True );
-
-  if( rules == NULL )
-  {
-    _xklLastErrorMsg = "Could not load rules";
-    return False;
-  }
-
-  if( !XkbRF_GetComponents( rules, &_xklVarDefs, &componentNames ) )
-  {
-    _xklLastErrorMsg = "Could not translate rules into components";
-    return False;
-  }
-
-  return True;
-}
-
-static void _XklConfigCleanAfterKbd(  )
-{
-  XkbRF_Free( rules, True );
-
-  if( locale != NULL )
-  {
-    free( locale );
-    locale = NULL;
-  }
-  if( _xklVarDefs.layout != NULL )
-  {
-    free( _xklVarDefs.layout );
-    _xklVarDefs.layout = NULL;
-  }
-  if( _xklVarDefs.options != NULL )
-  {
-    free( _xklVarDefs.options );
-    _xklVarDefs.options = NULL;
-  }
-}
-
-static void _XklApplyFun2XkbDesc( XkbDescPtr xkb, XkbDescModifierFunc fun,
-                                  void *userData, Bool activeInServer )
-{
-  int mask;
-  // XklDumpXkbDesc( "comp.xkb", xkb );
-  if( fun == NULL )
-    return;
-
-  if( activeInServer )
-  {
-    mask = ( *fun ) ( NULL, NULL );
-    if( mask == 0 )
-      return;
-    XkbGetUpdatedMap( _xklDpy, mask, xkb );
-    // XklDumpXkbDesc( "restored1.xkb", xkb );
-  }
-
-  mask = ( *fun ) ( xkb, userData );
-  if( activeInServer )
-  {
-    // XklDumpXkbDesc( "comp+.xkb", xkb );
-    XkbSetMap( _xklDpy, mask, xkb );
-    XSync( _xklDpy, False );
-
-    // XkbGetUpdatedMap( _xklDpy, XkbAllMapComponentsMask, xkb );
-    // XklDumpXkbDesc( "restored2.xkb", xkb );
+    //len = npc - pc;
+    len = strlen( pc );
+    //*ppc = ( char * ) strndup( pc, len );
+    *ppc = ( char * ) malloc( len + 1 );
+    if ( *ppc != NULL )
+      strcpy( *ppc, pc );
   }
 }
 
@@ -416,31 +354,42 @@ void XklConfigInit( void )
 void XklConfigTerm( void )
 {
   if( modelsXPath != NULL )
+  {
     xmlXPathFreeCompExpr( modelsXPath );
+    modelsXPath = NULL;
+  }
   if( layoutsXPath != NULL )
+  {
     xmlXPathFreeCompExpr( layoutsXPath );
+    layoutsXPath = NULL;
+  }
   if( optionGroupsXPath != NULL )
+  {
     xmlXPathFreeCompExpr( optionGroupsXPath );
+    optionGroupsXPath = NULL;
+  }
 }
 
-Bool XklConfigLoadRegistry( void )
+Bool XklConfigLoadRegistryFromFile( const char * fileName )
 {
-  theRegistry.doc = xmlParseFile( XML_CFG_PATH );
+  theRegistry.doc = xmlParseFile( fileName );
   if( theRegistry.doc == NULL )
   {
     theRegistry.xpathContext = NULL;
     _xklLastErrorMsg = "Could not parse XKB configuration registry";
   } else
     theRegistry.xpathContext = xmlXPathNewContext( theRegistry.doc );
-  return _XklConfigInitialized(  );
+  return _XklConfigRegistryIsInitialized(  );
 }
 
 void XklConfigFreeRegistry( void )
 {
-  if( _XklConfigInitialized(  ) )
+  if( _XklConfigRegistryIsInitialized(  ) )
   {
     xmlXPathFreeContext( theRegistry.xpathContext );
     xmlFreeDoc( theRegistry.doc );
+    theRegistry.xpathContext = NULL;
+    theRegistry.doc = NULL;
   }
 }
 
@@ -467,7 +416,7 @@ void XklConfigEnumOptionGroups( GroupProcessFunc func, void *userData )
   xmlXPathObjectPtr xpathObj;
   int i;
 
-  if( !_XklConfigInitialized(  ) )
+  if( !_XklConfigRegistryIsInitialized(  ) )
     return;
   xpathObj =
     xmlXPathCompiledEval( optionGroupsXPath, theRegistry.xpathContext );
@@ -564,117 +513,3 @@ Bool XklConfigFindOption( const char *optionGroupName,
       "[../configItem/name = '%s' and configItem/name = '%s']",
       optionGroupName, ptr, NULL );
 }
-
-Bool XklMultipleLayoutsSupported( void )
-{
-  struct stat buf;
-  return 0 == stat( MULTIPLE_LAYOUTS_CHECK_PATH, &buf );
-}
-
-Bool XklConfigActivate( const XklConfigRecPtr data, XkbDescModifierFunc fun,
-                        void *userData )
-{
-  Bool rv = False;
-  if( _XklConfigPrepareBeforeKbd( data ) )
-  {
-    XkbDescPtr xkb;
-    xkb =
-      XkbGetKeyboardByName( _xklDpy, XkbUseCoreKbd, &componentNames,
-                            XkbGBN_AllComponentsMask,
-                            XkbGBN_AllComponentsMask &
-                            ( ~XkbGBN_GeometryMask ), True );
-
-    _XklApplyFun2XkbDesc( xkb, fun, userData, True );
-
-    //!! Do I need to free it anywhere?
-    if( xkb != NULL )
-    {
-      //XklDumpXkbDesc( "config.xkb", xkb );
-#if 0
-      int i;
-      XklDebug( 150, "New model: [%s]\n", data->model );
-      XklDebug( 150, "New layout: [%s]\n", data->layout );
-      XklDebug( 150, "New variant: [%s]\n", data->variant );
-      for( i = data->numOptions; --i >= 0; )
-        XklDebug( 150, "New option[%d]: [%s]\n", i, data->options[i] );
-#endif
-      if( XklSetNamesProp( _XKB_RF_NAMES_PROP_ATOM, RULES_FILE, data ) )
-        rv = True;
-      else
-        _xklLastErrorMsg = "Could not set names property";
-      XkbFreeKeyboard( xkb, XkbAllComponentsMask, True );
-    } else
-    {
-      _xklLastErrorMsg = "Could not load keyboard description";
-    }
-  }
-  _XklConfigCleanAfterKbd(  );
-  return rv;
-}
-
-int XklSetKeyAsSwitcher( XkbDescPtr kbd, void *userData )
-{
-  if( kbd != NULL )
-  {
-    XkbClientMapPtr map = kbd->map;
-    if( map != NULL )
-    {
-      KeySym keysym = ( KeySym ) userData;
-      KeySym *psym = map->syms;
-      int symno;
-
-      for( symno = map->num_syms; --symno >= 0; psym++ )
-      {
-        if( *psym == keysym )
-        {
-          XklDebug( 160, "Changing %s to %s at %d\n",
-                    XKeysymToString( *psym ),
-                    XKeysymToString( XK_ISO_Next_Group ), psym - map->syms );
-          *psym = XK_ISO_Next_Group;
-          break;
-        }
-      }
-    } else
-      XklDebug( 160, "No client map in the keyboard description?\n" );
-  }
-  return XkbKeySymsMask | XkbKeyTypesMask | XkbKeyActionsMask;
-}
-
-Bool XklConfigWriteXKMFile( const char *fileName, const XklConfigRecPtr data,
-                            XkbDescModifierFunc fun, void *userData )
-{
-  Bool rv = False;
-
-  FILE *output = fopen( fileName, "w" );
-  XkbFileInfo dumpInfo;
-
-  if( output == NULL )
-  {
-    _xklLastErrorMsg = "Could not open the XKB file";
-    return False;
-  }
-
-  if( _XklConfigPrepareBeforeKbd( data ) )
-  {
-    XkbDescPtr xkb;
-    xkb =
-      XkbGetKeyboardByName( _xklDpy, XkbUseCoreKbd, &componentNames,
-                            XkbGBN_AllComponentsMask,
-                            XkbGBN_AllComponentsMask &
-                            ( ~XkbGBN_GeometryMask ), False );
-    if( xkb != NULL )
-    {
-      _XklApplyFun2XkbDesc( xkb, fun, userData, False );
-
-      dumpInfo.defined = 0;
-      dumpInfo.xkb = xkb;
-      dumpInfo.type = XkmKeymapFile;
-      rv = XkbWriteXKMFile( output, &dumpInfo );
-      XkbFreeKeyboard( xkb, XkbGBN_AllComponentsMask, True );
-    } else
-      _xklLastErrorMsg = "Could not load keyboard description";
-  }
-  _XklConfigCleanAfterKbd(  );
-  fclose( output );
-  return rv;
-}
